import json
import os.path
from concurrent.futures import ProcessPoolExecutor
from multiprocessing import Queue, Process

import aiohttp_cors
import cv2

from utils.helper_functions import *
from utils.models import make_and_save_model
from utils.progress_report_socket import *
from utils.weighted_binary_crossentropy import WeightedBinaryCrossentropy

app = web.Application()
pool = ProcessPoolExecutor(max_workers=2)
queue = None


async def main(request):
    """
    Entry point function of the API, this function is called when client sends a request to the API.
    Provides the functionality described in detail in the Bachelor thesis "Tool for data pre-processing and iterative learning of neural networks".
    In short, this function preprocesses images, trains a neural network on these images and uses the neural network for inference.
    :param request: received JSON from client
    :return: response for the client
    """

    input = await request.text()
    json_input = json.loads(input)

    # check necessary parameters
    if 'train' in json_input:
        if 'inputs' not in json_input['train']:
            return error_response('missing required field inputs')
        if 'gts' not in json_input['train']:
            return error_response('missing required field gts')
        if 'validation_ratio' not in json_input['train']:
            return error_response('missing required field validation_ratio')
        if 'epoch_count' not in json_input['train']:
            return error_response('missing required field epoch_count')
        if 'preprocessing_params' not in json_input:
            return error_response('missing preprocessing parameters')
        if 'patch_size' not in json_input['preprocessing_params']:
            return error_response('missing preprocessing parameter patch size')
        if 'batch_size' not in json_input['preprocessing_params']:
            return error_response('missing preprocessing parameter batch size')
        if 'overlap' not in json_input['preprocessing_params']:
            return error_response('missing preprocessing parameter overlap')

        # load images
        train_x = []
        train_y = []
        for data, label in zip(json_input['train']['inputs'], json_input['train']['gts']):
            image = [cv2.imread(layer) for layer in data]
            if any(i is None for i in image):
                print(data)
                return error_response('invalid image file name')
            image = np.concatenate(image, axis=2)
            train_x.append(image)
            mask = cv2.imread(label, 0)
            if mask is None:
                print(label)
                return error_response('invalid mask file name')
            mask = list(map(lambda row: list(map(lambda x: [float(x)], row)), mask))
            train_y.append(mask)

        train_x, train_y = preprocess(json_input, train_x, train_y)

        # load model (or create new) and prepare training parameters
        if not os.path.isdir(json_input['dataset_id'] + '.model'):
            make_and_save_model(json_input['dataset_id'] + '.model')
        model = tf.keras.models.load_model(json_input['dataset_id'] + '.model',
                                           custom_objects={'WeightedBinaryCrossentropy': WeightedBinaryCrossentropy})
        batch_size = json_input['preprocessing_params']['batch_size']
        n_epochs = json_input['train']['epoch_count']
        validation_ratio = json_input['train']['validation_ratio']
        early_stopping = tf.keras.callbacks.EarlyStopping(patience=10, restore_best_weights=True)
        weights = calc_class_weights(train_y)

        # compile and train model
        model.compile(optimizer='adam',
                      loss=WeightedBinaryCrossentropy(weights),
                      metrics=[tf.keras.metrics.BinaryAccuracy()])
        model.fit(x=train_x, y=train_y,
                  batch_size=batch_size,
                  epochs=n_epochs,
                  validation_split=validation_ratio,
                  shuffle=True,
                  callbacks=[SendProgressThroughWebSocket(json_input['dataset_id'], n_epochs, queue), early_stopping],
                  verbose=0)
        model.save(json_input['dataset_id'] + '.model')
    if 'inference' in json_input:
        # check necessary parameters
        if 'inputs' not in json_input['inference']:
            return error_response('missing required field inputs')
        if 'outputs' not in json_input['inference']:
            return error_response('missing required field outputs')
        if len(json_input['inference']['inputs']) != len(json_input['inference']['outputs']):
            return error_response('lengths of inputs and outputs do not match')

        # load model
        if not os.path.isdir(json_input['dataset_id'] + '.model'):
            return error_response('no trained model found for given dataset')
        model = tf.keras.models.load_model(json_input['dataset_id'] + '.model',
                                           custom_objects={'WeightedBinaryCrossentropy': WeightedBinaryCrossentropy})

        # load and normalize images and save masks
        images = []
        for data in json_input['inference']['inputs']:
            image = [cv2.imread(layer) for layer in data]
            if any(i is None for i in image):
                return error_response('invalid image filename for inference')
            images.append(image)

        for image, filename in zip(images, json_input['inference']['outputs']):
            infer(filename, image, model, json_input['dataset_id'])

    return web.json_response({'status': 'OK'})


def infer(filename, image, model, dataset_id):
    """
    Infers (creates using a neural network) a mask and saves it
    :param filename: to store the inferred mask
    :param image: to infer
    :param model: to use for inference
    :param dataset_id: needed for normalisation
    """

    image = np.concatenate(image, axis=2)
    image = np.array(tf.expand_dims(image, axis=0))
    image2 = normalise(dataset_id, image)

    # perform annotation
    mask = model.predict(image2)
    mask = mask.squeeze()
    mask = tf.slice(mask, [0, 0], image.shape[1:3])

    # convert to 0s and 1s
    mask = np.array(tf.round(mask)) * 255

    # save mask
    cv2.imwrite(filename, mask)


def preprocess(json_input, train_x, train_y):
    """
    Preprocess images and masks (normalisation, patching, filtering, shuffling)
    :param json_input: the JSON received in the request
    :param train_x: images
    :param train_y: (true) masks
    :return: preprocessed images and masks
    """

    # normalize labels (to 0s and 1s)
    train_y = map(lambda mask: list(map(lambda x: list(map(lambda y: [0.] if y[0] < 100 else [1.],
                                                           x)), mask)), train_y)

    # pad images to achieve same dimensions (currently not necessary)
    # train_x = map(lambda image: tf.image.pad_to_bounding_box(image, 0, 0, 1000, 1000), train_x)
    # train_y = map(lambda mask: tf.image.pad_to_bounding_box(mask, 0, 0, 1000, 1000), train_y)
    train_x = np.array(list(train_x))
    train_y = np.array(list(train_y))

    # normalize images
    train_x = normalise(json_input['dataset_id'], train_x)

    # create patches
    train_x = patch_up(train_x, json_input, 3)
    train_y = patch_up(train_y, json_input, 1)

    # remove empty patches
    train = zip(train_x, train_y)
    train = list(filter(lambda img_lbl: img_lbl[1].sum() > 20, train))
    train_x = np.array([x for x, _ in train])
    train_y = np.array([x for _, x in train])

    # shuffle (seeds need to be the same to shuffle images and masks the same way)
    tf.random.set_seed(5)
    train_x = np.array(tf.random.shuffle(train_x, seed=5))
    tf.random.set_seed(5)
    train_y = np.array(tf.random.shuffle(train_y, seed=5))
    return train_x, train_y


def normalise(dataset_id, images):
    """
    Normalise images based on mean and standard deviation (images - mean)/standard_deviation,
     calculates them during first iteration and stores them for later uses, loads them from file during later iterations
    :param dataset_id: used for storing (and loading) of calculated mean and standard deviation, in the filename
    :param images: to be normalised
    :return: normalised images
    """

    images = images.astype('float64')
    if os.path.isfile(dataset_id + '_mean.npy'):
        with open(dataset_id + '_mean.npy', 'rb') as m:
            mean = np.load(m)
    else:
        mean = images.mean(axis=(0, 1, 2), keepdims=True)
        with open(dataset_id + '_mean.npy', 'wb') as m:
            np.save(m, mean)

    if os.path.isfile(dataset_id + '_std.npy'):
        with open(dataset_id + '_std.npy', 'rb') as s:
            std = np.load(s)
    else:
        std = np.fmax(images.std(axis=(0, 1, 2, 3), keepdims=True), np.divide(1., np.sqrt(images.size)))
        with open(dataset_id + '_std.npy', 'wb') as s:
            np.save(s, std)

    images = (images - mean) / std
    return images


def prepare_main_server(q):
    """
    cors setup and run API on port 5000
    :param q: multiprocessing queue used for communication between API and progress reporter
    """

    global queue
    queue = q
    cors = aiohttp_cors.setup(app, defaults={
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=True,
            expose_headers="*",
            allow_headers="*",
        )
    })
    cors.add(app.router.add_post('/', main))
    cors.add(app.router.add_post('/api/iteration', main))
    loop = asyncio.get_event_loop()
    loop.create_task(web._run_app(app, port=5000))
    loop.run_forever()


def prepare_progress_reporter(q):
    """
    cors setup and run progress reporter on port 5001
    :param q: multiprocessing queue used for communication between API and progress reporter
    """

    cors = aiohttp_cors.setup(progress_reporter, defaults={
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=True,
            expose_headers="*",
            allow_headers="*",
        )
    })
    loop = asyncio.new_event_loop()
    loop.create_task(web._run_app(progress_reporter, port=5001))
    loop.create_task(send_progress(q))
    loop.run_forever()


# start the API and the progress reporter as separate processes
if __name__ == '__main__':
    queue = Queue()
    p1 = Process(target=prepare_main_server, args=(queue,))
    p1.start()
    p2 = Process(target=prepare_progress_reporter, args=(queue,))
    p2.start()
    p1.join()
    p2.join()
