import json
import os
import time

import cv2
import matplotlib.pyplot as plt
import numpy as np
import requests
import sklearn.model_selection
import tensorflow as tf
from pathlib import Path

import API
from utils.helper_functions import calc_class_weights


def save_comparison(display_list, name):
    """
    Plots the comparison (of image, its true mask and generated mask) and saves it
    :param display_list: image, true mask, generated mask (numpy arrays)
    :param name: filename
    """
    plt.figure(figsize=(15, 15))

    title = ['Input Image', 'True Mask', 'Predicted Mask']

    for i in range(len(display_list)):
        plt.subplot(1, len(display_list), i + 1)
        plt.title(title[i])
        plt.imshow(tf.keras.preprocessing.image.array_to_img(display_list[i]))
        plt.axis('off')
    plt.savefig(name)


class ExperimentExecutor:
    """
    Class for carrying out experiments with the API, configurable with a JSON file
    """

    def __init__(self, config_filename) -> None:
        """
        Constructor of Experiment Executor, prepares training and testing images and masks based on the configuration JSON file
        :param config_filename: file name of configuration file
        """

        super().__init__()
        with open(config_filename) as JSON:
            self.experiments = json.load(JSON)
            location = self.experiments['data']['data_location']
            image_name = self.experiments['data']['image_format']
            mask_name = self.experiments['data']['mask_format']
            starting_image = self.experiments['data']['image_start']
            number_of_images = self.experiments['data']['number_of_images']
            image_range = range(starting_image, starting_image + number_of_images)
            images = [[os.path.join(location, image_name.format(i))] for i in image_range]
            masks = [os.path.join(location, mask_name.format(i)) for i in image_range]
            self.training_images, self.testing_images, self.training_masks, self.testing_masks = sklearn.model_selection.train_test_split(
                images, masks, test_size=self.experiments['data']['test_size'], random_state=1)

    def _make_request_and_save_sidebyside(self, dataset_id, num_images, iteration_number):
        """
        Sends an inference request to the API, then makes a side by side comparison of image with its true and generated masks
        :param dataset_id: name of current experiment (used as the dataset id)
        :param num_images: number of images to make comparisons of
        :param iteration_number: used in the filename of the comparison
        """

        req = {
            'dataset_id': dataset_id,
            'inference': {
                'inputs': self.testing_images[:num_images],
                'outputs': [os.path.join(self.experiments['result_directory'], dataset_id,
                                         f'test{i}iteration{iteration_number}.png') for i in range(num_images)]
            }
        }
        response = requests.post(self.experiments['API_address'], json=req)
        for i, t, m in zip(self.testing_images[:num_images], self.testing_masks[:num_images], range(num_images)):
            image = cv2.imread(i[0])
            true = cv2.imread(t)
            mask = cv2.imread(os.path.join(self.experiments['result_directory'], dataset_id,
                                           f'test{m}iteration{iteration_number}.png'))
            save_comparison([image, true, mask],
                            os.path.join(self.experiments['result_directory'], dataset_id,
                                         f'comparison{m}iteration{iteration_number}.png'))

    def _evaluate_model(self, dataset_id):
        """
        Sends inference request to the API, then compares the generated masks to the true masks using the custom
        loss function (the same the API uses) and Jaccard index, the results are saved to a file
        :param dataset_id: name of current experiment (used as the dataset id)
        :return evaluation results as string
        """

        req = {
            'dataset_id': dataset_id,
            'inference': {
                'inputs': self.testing_images,
                'outputs': list(map(lambda x: x[0][:-3] + 'testmask.png', self.testing_images))
            }
        }
        response = requests.post(self.experiments['API_address'], json=req)
        true_masks = np.array(
            [np.expand_dims(cv2.imread(i, cv2.IMREAD_GRAYSCALE), axis=-1) / 255 for i in self.testing_masks])
        predicted_masks = np.array([np.expand_dims(cv2.imread(i, cv2.IMREAD_GRAYSCALE), axis=-1) / 255 for i in
                                    list(map(lambda x: x[0][:-3] + 'testmask.png', self.testing_images))])
        weights = calc_class_weights(true_masks)
        loss = API.WeightedBinaryCrossentropy(weights).call(true_masks, predicted_masks).numpy()
        eval_res = dict()
        eval_res['loss'] = loss
        jaccard = tf.keras.metrics.MeanIoU(num_classes=2)
        jaccard.update_state(true_masks, predicted_masks)
        eval_res['Jaccard index'] = jaccard.result().numpy()
        # print('eval:', eval_res)
        return f'eval: {eval_res}\n'

    def _learning_request(self, name, number=None, epochs=40):
        """
        Sends a learning request to the API, takes parameters from configuration JSON file
        :param name: name of experiment, used as dataset id
        :param number: how many images (and masks) are to be used for training
        :param epochs: number of epochs the training will run for
        """

        req = {
            'dataset_id': name,
            'train': {
                'validation_ratio': self.experiments['training_parameters']['validation_ratio'],
                'epoch_count': epochs,
                'gts': self.training_masks[:number + 1] if number is not None else self.training_masks,
                'inputs': self.training_images[:number + 1] if number is not None else self.training_images
            },
            'preprocessing_params': {
                'patch_size': self.experiments['training_parameters']['patch_size'],
                'overlap': self.experiments['training_parameters']['overlap'],
                'batch_size': self.experiments['training_parameters']['batch_size']
            },
        }
        response = requests.post(self.experiments['API_address'], json=req)
        # print(response.json())

    def _perform_experiment(self, details: dict):
        """
        Carries out an experiment, as defined in the configuration file.
        Evaluates the model after each iteration, possibly also creates comparisons, measures training time and saves results to file.
        :param details: of the experiment being carried out
        """

        Path(os.path.join(self.experiments['result_directory'], details['name'])).mkdir(parents=True, exist_ok=True)
        # print(details['name'])
        result_message = details['name'] + '\n'
        total_time = 0
        for i in range(details['num_of_iterations']):
            # print('iteration', i)
            result_message += f'iteration {i}\n'
            number_of_images = None if i + 1 == details['num_of_iterations'] and details.get(
                'last_iteration_all_images', False) else details['initial_number_of_images'] + i * details[
                'iteration_image_add'] + (pow(i, 2) // 2 - i // 2) * details['iteration_image_add_change']
            number_of_epochs = details['initial_number_of_epochs'] + i * details.get('iteration_epochs_add', 0) + (
                    pow(i, 2) // 2 - i // 2) * details.get('iteration_epochs_add_change', 0)
            start = time.time()
            self._learning_request(details['name'], number_of_images, number_of_epochs)
            end = time.time()
            total_time += end - start
            result_message += self._evaluate_model(details['name'])
            if 'comparison_images' in details:
                self._make_request_and_save_sidebyside(details['name'], details['comparison_images'], i)
        # print(f'Took {total_time} seconds')
        result_message += f'Took {total_time} seconds\n'
        with open(os.path.join(self.experiments['result_directory'], details['name'], 'results.txt'), 'a') as file:
            file.write(result_message)

    def run_experiments(self):
        """
        Runs all experiments.
        """

        for experiment in self.experiments['experiments']:
            self._perform_experiment(experiment)


# performs experiments defined in JSON configuration file
if __name__ == '__main__':
    executor = ExperimentExecutor('experiments_config.json')
    executor.run_experiments()
