import numpy as np
import tensorflow as tf


class WeightedBinaryCrossentropy(tf.keras.losses.Loss):
    """
    Custom loss function used by the API in the training of neural network.
    Takes into consideration the weights of labels.
    Used to counter balance imbalanced data
    """

    def __init__(self, weights=None, reduction=tf.keras.losses.Reduction.NONE, name=None):
        super().__init__(reduction, name)
        if weights is None:
            weights = []
        self.weights = weights

    def call(self, y_true, y_pred):
        """
        This function is used during training to calculate loss
        :param y_true: ground truth (true masks)
        :param y_pred: predicted labels (masks)
        :return: weighted loss
        """

        y_true = tf.cast(y_true, dtype=np.float64)
        y_pred = tf.cast(y_pred, dtype=np.float64)

        weights_v = tf.where(tf.equal(y_true, 1.), self.weights[1], self.weights[0])
        weights_v = tf.cast(weights_v, dtype=np.float64)
        ce = tf.keras.backend.binary_crossentropy(y_true, y_pred, from_logits=False)

        loss = tf.reduce_mean(tf.multiply(ce, weights_v))

        return loss
